jobspec_nextgen = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "http://github.com/flux-framework/rfc/tree/master/data/spec_24/schema.json",
    "title": "jobspec-01",
    "description": "JobSpec the Next Generation",
    "type": "object",
    # The only required thing is a version. Tasks and groups can be defined.
    # If neither is, we essentially do nothing.
    "required": ["version"],
    "properties": {
        # Name for the entire jobspec is optional
        "name": {"type": "string"},
        # This is not a flux JobSpec, and we start at v1
        "version": {
            "description": "the jobspec version",
            "type": "integer",
            "enum": [1],
        },
        "requires": {
            "type": "object",
            "patternProperties": {
                "^([a-z]|[|]|&|[0-9]+)+$": {
                    "type": "array",
                    "items": {"type": "object"},
                },
            },
        },
        "resources": {
            "type": "object",
            "patternProperties": {
                "^([a-z]|[|]|&|[0-9]+)+$": {"$ref": "#/definitions/resources"},
            },
        },
        # The top level jobspec has groups and tasks
        # Groups are "flux batch"
        "groups": {"type": "array", "items": {"$ref": "#/definitions/group"}},
        # Tasks are one or more named tasks
        # Tasks are "flux submit" on the level they are defined
        "tasks": {"$ref": "#/definitions/tasks"},
        # Attributes can eventually be given to the jobs
        "attributes": {"$ref": "#/definitions/attributes"},
        "additionalProperties": False,
    },
    "definitions": {
        "attributes": {
            "description": "system, parameter, and user attributes",
            "type": "object",
            "properties": {
                "duration": {"type": "number", "minimum": 0},
                "cwd": {"type": "string"},
                "environment": {"type": "object"},
            },
        },
        "resources": {
            "description": "requested resources",
            "type": "object",
            "required": ["type"],
            "properties": {
                "type": {"type": "string"},
                # Count is only required when below a slot
                "count": {"type": "integer", "minimum": 1},
                "requires": {
                    "type": "array",
                    "items": {"type": "object"},
                },
                "attributes": {"$ref": "#/definitions/attributes"},
                "schedule": {"type": "boolean"},
                "with": {
                    "type": "array",
                    "minItems": 1,
                    "items": {"$ref": "#/definitions/resources"},
                },
            },
        },
        "steps": {
            "type": ["array"],
            "items": {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string",
                        "enum": ["stage"],
                    },
                },
                "required": ["name"],
            },
        },
        "tasks": {
            "description": "tasks configuration",
            "type": "array",
            # If no slot is defined, it's implied to be at the top level (the node)
            "items": {
                "type": "object",
                "properties": {
                    # These are task level items that over-ride global
                    # Resources in a task can be traditional OR a string reference
                    "resources": {"type": "string"},
                    # A task can reference another group (a flux batch)
                    "group": {"type": "string"},
                    # If the task is run locally on the level it is currently at.
                    "local": {"type": "boolean"},
                    # Requires is a reference to one or more requirement blocks
                    # For example, users space subsystems
                    "requires": {"type": "string"},
                    # Name only is needed to reference the task elsewhere
                    "name": {"type": "string"},
                    "depends_on": {"type": "array", "items": {"type": "string"}},
                    # How many of this task are to be run?
                    "replicas": {"type": "number", "minimum": 1, "default": 1},
                    # A command can be a string or a list of strings
                    "command": {
                        "type": ["string", "array"],
                        "minItems": 1,
                        "items": {"type": "string"},
                    },
                    # Custom logic for the transformer
                    "steps": {"$ref": "#/definitions/steps"},
                },
            },
        },
        "group": {
            "description": "group of tasks (batch)",
            "type": "object",
            # If no slot is defined, it's implied to be at the top level (the node)
            "properties": {
                # Name only is needed to reference the group elsewhere
                "name": {"type": "string"},
                "resources": {"type": "string"},
                "depends_on": {"type": "array", "items": {"type": "string"}},
                # Tasks for the group
                "tasks": {"$ref": "#/definitions/tasks"},
                "groups": {"type": "array", "items": {"$ref": "#/definitions/group"}},
            },
            "additionalProperties": False,
        },
    },
}



#!/usr/bin/env python

import sys

from jobspec.subsystem import get_subsystem_registry


def main(args, _):
    """
    Determine if a jobspec can be satsified by local resources.
    This is a fairly simple (flat) check.
    """
    registry = get_subsystem_registry(args.sdir)
    is_satisfied = registry.satisfied(args.jobspec, ignore_missing=not args.require_all)
    sys.exit(0 if is_satisfied else -1)



from jobspec.logger import logger


class Result:
    """
    Helper class to encompass a result
    """

    def __init__(self, out=None, prefix="   ", has_depends_on=False):
        self.out = out or ""
        self.lines = []
        self.prefix = prefix
        self.has_depends_on = has_depends_on

    def add_debug_line(self, line):
        self.lines.append(line)

    def print_extra(self):
        for line in self.lines:
            logger.debug(self.prefix + line)



import importlib
import pkgutil

from jobspec.transformer import plugins as in_tree


class TransformerRegistry:
    """
    A registry of transformers - jobspec steps to workload manager setup.
    """

    plugin_class = "Transformer"
    module_prefix = "jobspec_"

    def __init__(self):
        self.discover()

    def get_plugin_names(self):
        return list(self.plugins)

    def is_installed(self, name):
        return name in self.plugins

    def get_plugin(self, name):
        plugin = self.plugins.get(name)
        if plugin is None:
            raise ValueError(f"Plugin {name} is not known.")
        return plugin

    def add_arguments(self, subparser):
        """
        Add subparser and arguments for each plugin
        """
        for _, plugin in self.plugins.items():
            plugin.add_arguments(subparser)

    def discover(self):
        """
        Discover and register plugins with name jobspec-<name>
        """
        self.plugins = {}
        for moduleinfo in pkgutil.iter_modules():
            if not moduleinfo.ispkg or not moduleinfo.name.startswith(self.module_prefix):
                continue
            module = importlib.import_module(moduleinfo.name)
            self.register(moduleinfo.name, module)

        # These are internally provided
        for module in in_tree:
            self.register_in_tree(module)

    def register_in_tree(self, plugin):
        """
        Register a new in tree plugin.

        The main difference here is that in tree plugins
        are already loaded.
        """
        if not hasattr(plugin, "name"):
            raise ValueError(f"Plugin {plugin} is missing a name")
        name = plugin.name
        if name in self.plugins:
            return
        self._validate_plugin(plugin, name)
        self.plugins[name] = plugin

    def register(self, name, plugin):
        """
        Register a new plugin.
        """
        if name in self.plugins:
            return
        self.validate_plugin(name, plugin)

        # Python modules use "_" instead of "-"
        name = name.removeprefix(self.module_prefix).replace("_", "-")
        self.plugins[name] = self.load_plugin(name, plugin)

    def load_plugin(self, name, module):
        """
        Load a plugin
        """
        cls = getattr(module, self.plugin_class)
        return cls(name)

    def validate_plugin(self, name, module):
        """
        Validate a plugin.
        """
        invalid = f"Plugin {name} is not valid"

        # Plugin must be defined!
        if not hasattr(module, self.plugin_class):
            raise ValueError(f"{invalid}, missing {self.plugin_class} to import")

        # Defaults must also be defined
        if not hasattr(module, "defaults"):
            raise ValueError(f"{invalid}, missing 'defaults' submodule.")

        cls = getattr(module, self.plugin_class)
        self._validate_plugin(cls, name)

    def _validate_plugin(self, cls, name):
        """
        Shared validation steps.
        """
        invalid = f"Plugin {name} is not valid"

        # Class attributes
        for attr in ["description", "name"]:
            if not hasattr(cls, attr):
                raise ValueError(f"{invalid}, missing '{attr}' attribute")
            if not getattr(cls, attr, None):
                raise ValueError(f"{invalid},'{attr}' attribute is not defined")



import json
import os

import yaml

import jobspec.utils as utils


class ResourceBase:
    def __init__(self, data):
        """
        Interact with loaded resources.
        """
        self.data = data or {}

    def to_str(self):
        """
        Convert to string
        """
        return json.dumps(self.data)

    def to_yaml(self):
        """
        Dump to yaml string
        """
        return yaml.dump(self.data)

    def get(self, name, default=None):
        """
        Wrapper to the self.data
        """
        if not self.data:
            return default
        return self.data.get(name, default)

    def update(self, attrs):
        """
        Update with new key value pairs
        """
        if not attrs:
            return
        self.data.update(attrs)

    def load(self, filename):
        """
        Load the jobspec (from file or as string)
        """
        # Case 1: given a raw filename
        if isinstance(filename, str) and os.path.exists(filename):
            self.filename = os.path.abspath(filename)

            try:
                self.data = utils.read_json(self.filename)
            except:
                self.data = utils.read_yaml(self.filename)

        # Case 2: jobspec as dict (that we just want to validate)
        elif isinstance(filename, dict):
            self.data = filename
        # Case 3: jobspec as string for json or yaml
        else:
            try:
                self.data = json.loads(filename)
            except:
                self.data = yaml.load(filename, Loader=yaml.SafeLoader)

        # Case 4: wtf are you giving me? :X
        if not self.data:
            raise ValueError("Unrecognized input format for {self.__class__.__name__}.")



import os
import sqlite3

import jobspec.core as core
import jobspec.subsystem.queries as queries
import jobspec.utils as utils
from jobspec.logger import LogColors, logger


class SubsystemRegistry:
    """
    A subsystem registry has (and loads) one or more subsystems.

    Right now we use an in memory sqlite database since it's
    efficient.
    """

    def __init__(self, path):
        self.systems = {}
        self.conn = sqlite3.connect(":memory:")
        self.create_tables()
        self.load(path)

    def __exit__(self):
        self.close()

    def close(self):
        self.conn.close()

    def create_tables(self):
        """
        Create tables for subsytems, nodes, edges.

        Note that I'm flattening the graph, so edges become attributes for
        nodes so it's easy to query. This is a reasonable first shot over
        implementing an actual graph database.
        """
        cursor = self.conn.cursor()

        # Only save metadata we absolutely need
        # Note I'm not saving edges because we don't use
        # them for anything - we are going to parse them
        # into node attributes instead.
        create_sql = [
            queries.create_subsystem_sql,
            queries.create_nodes_sql,
            queries.create_attributes_sql,
        ]
        for sql in create_sql:
            cursor.execute(sql)
        self.conn.commit()

    def load(self, path):
        """
        Load a group of subsystem files, typically json JGF.
        """
        if not os.path.exists(path):
            raise ValueError(f"User subsystem directory {path} does not exist.")
        files = utils.recursive_find(path, "[.]json")
        if not files:
            raise ValueError(f"There are no subsystem files in {path}")
        for filename in files:
            new_subsystem = Subsystem(filename)
            self.load_subsystem(new_subsystem)

    def load_subsystem(self, subsystem):
        """
        Load a new subsystem to the memory database
        """
        cursor = self.conn.cursor()

        # Create the subsystem - it should error if already exists
        values = f"('{subsystem.name}')"
        fields = '("name")'
        statement = f"INSERT INTO subsystems {fields} VALUES {values}"
        logger.debug(statement)
        cursor.execute(statement)
        self.conn.commit()

        # These are fields to insert a node and attributes
        node_fields = '("subsystem", "label", "type", "basename", "name", "id")'

        # First create all nodes
        for nid, node in subsystem.graph["nodes"].items():
            typ = node["metadata"]["type"]
            basename = node["metadata"]["basename"]
            name = node["metadata"]["name"]
            id = node["metadata"]["id"]
            node_values = f"('{subsystem.name}', '{nid}', '{typ}', '{basename}', '{name}', '{id}')"
            statement = f"INSERT INTO nodes {node_fields} VALUES {node_values}"
            logger.debug(statement)
            cursor.execute(statement)

        # Commit transaction
        self.conn.commit()
        attr_fields = '("node", "name", "value")'

        # Now all attributes
        for nid, node in subsystem.graph["nodes"].items():
            for key, value in node["metadata"].get("attributes", {}).items():
                attr_values = f"('{nid}', '{key}', '{value}')"
                statement = f"INSERT INTO attributes {attr_fields} VALUES {attr_values}"
                logger.debug(statement)
                cursor.execute(statement)

        # Note that we aren't doing anything with edges.
        # We are going to query for nodes we need directly. This assumes
        # that a software environment would be a global, resource-wide thing.
        self.conn.commit()

    def get_subsystem_node_type(self, subsys, match):
        """
        Get nodes of a specific type under a subsystem
        """
        statement = f"SELECT label from nodes WHERE subsystem = '{subsys}' AND type = '{match}';"
        labels = self.query(statement)
        return [f"'{x[0]}'" for x in labels]

    def find_node_attributes(self, attribute, value, labels=None):
        """
        Given a list of node labels, find children (attributes)
        that have a specific key/value.
        """
        # These will typically be node ids, but we need to filter attribute
        listing = "(%s)" % ",".join(labels)
        if labels is not None:
            statement = f"SELECT * from attributes WHERE name = '{attribute}' AND value = '{value}' and node IN {listing};"
        else:
            statement = (
                f"SELECT * from attributes WHERE name = '{attribute}' AND value = '{value}';"
            )
        return self.query(statement)

    def query(self, statement):
        """
        Issue a query to the database, returning fetchall.
        """
        cursor = self.conn.cursor()
        printed = statement

        # Don't overwhelm the output!
        if len(printed) > 150:
            printed = printed[:150] + "..."
        logger.info(printed)

        cursor.execute(statement)
        self.conn.commit()
        return cursor.fetchall()

    def satisfied(self, jobspec, ignore_missing=True):
        """
        Determine if a jobspec is satisfied by user-space subsystems.

        If ignore_missing is true, we assume missing the data does not
        disqualify the job.
        """
        js = core.Jobspec(jobspec)

        # We don't care about the association with tasks - the requires must be met
        # We could optimize this to be fewer queries, but it's likely trivial for now
        for _, requires in js.get("requires", {}).items():
            for item in requires:
                # If this returns None, ignore_missing is True and we ignore/continue
                subsys = self.get_item_subsystem(item, ignore_missing)
                if not subsys:
                    continue

                # Right now just require either:
                # 1. type, match, without attribute (no query to attribute table)
                # 2. type, match, attribute, value (query to nodes and attribute table)
                # 3. attribute and value (query only to attribute table)
                attribute = item.get("attribute")
                value = item.get("value")
                field = item["field"]
                match = item["match"]

                # We are being strict now and enforcing that field == type
                # We could support more, but would need to add them to the database
                # custom attributes should go under attributes
                if field != "type":
                    logger.warning(
                        f'Item {item} is searching for field other than "type," not supported yet.'
                    )
                    continue

                # We need at least a set of either
                if not all([attribute, value]) or not all([field, match]):
                    logger.warning(
                        f"Item {item} is missing 'field' and/or 'match' and cannot be assessed."
                    )
                    continue

                # 2. type, match, attribute, value (query to nodes and attribute table)
                if all([attribute, value, match]):
                    # "Get nodes in subsystem X of this type"
                    labels = self.get_subsystem_node_type(subsys, match)
                    if not labels:
                        print(
                            f"{LogColors.OKBLUE}{js.name}{LogColors.ENDC} {LogColors.RED}NOT OK{LogColors.ENDC}"
                        )
                        return False

                    # "Get attribute key values associated with any of these nodes"
                    # Here we get back matches that mean we are good (satisfied)
                    matches = self.find_node_attributes(attribute, value, labels)
                    if not matches:
                        print(
                            f"{LogColors.OKBLUE}{js.name}{LogColors.ENDC} {LogColors.RED}NOT OK{LogColors.ENDC}"
                        )
                        return False

                # 3. attribute and value (query only to attribute table)
                elif all([attribute, value]):
                    matches = self.find_node_attributes(attribute, value)
                    if not matches:
                        print(
                            f"{LogColors.OKBLUE}{js.name}{LogColors.ENDC} {LogColors.RED}NOT OK{LogColors.ENDC}"
                        )
                        return False

                # 1. type, match, without attribute (no query to attribute table)
                elif match is not None:
                    labels = self.get_subsystem_node_type(subsys, match)
                    if not labels:
                        print(
                            f"{LogColors.OKBLUE}{js.name}{LogColors.ENDC} {LogColors.RED}NOT OK{LogColors.ENDC}"
                        )
                        return False

        print(f"{LogColors.OKBLUE}{js.name}{LogColors.ENDC} {LogColors.OKGREEN}OK{LogColors.ENDC}")
        return True

    def get_item_subsystem(self, item, ignore_missing=True):
        """
        Get the subsystem for an item
        """
        # Check 1: the item provides the name of a subsystem
        group = item.get("name")
        if not group:
            msg = f"User subsystem requirement {item} is missing a subsystem name."
            if ignore_missing:
                logger.warning(msg)
            else:
                raise ValueError(msg)

        # Check 2: the subsystem exists in our database
        statement = f"SELECT name from subsystems WHERE name = '{group}';"
        match = self.query(statement)
        if not match and not ignore_missing:
            raise ValueError(f"User subsystem {group} is not known, and is required.")

        return group


class Subsystem:
    def __init__(self, filename):
        """
        Load a single subsystem
        """
        self._name = None
        self.load(filename)

    def load(self, filename):
        """
        Load a subsystem file, ensuring it exists.
        """
        # Derive the subsystem name from the filepath
        basename = os.path.basename(filename)
        self.data = utils.read_json(filename)

        if "graph" not in self.data:
            raise ValueError(f"Subsystem from {basename} is missing a graph")

        # Nodes are required (edges are not)
        if "nodes" not in self.graph or not self.graph["nodes"]:
            raise ValueError(f"Subsystem from {basename} is missing nodes")

    @property
    def graph(self):
        """
        Return the graph, which is required to exist and be populated to load.
        """
        return self.data["graph"]

    @property
    def name(self):
        """
        The subsystem name is typically the top level type of resource from the first
        node in the graph. We also could have a metadata feature.
        """
        if self._name is not None:
            return self._name
        # The subsystem won't load without nodes, so we have them
        for nid, node in self.graph["nodes"].items():
            # This is a fallback to invalid metadata (which should not happen)
            if "metadata" not in node or "type" not in node["metadata"]:
                self._name = nid.replace("0", "")
                return self._name

            self._name = node["metadata"]["type"]
            return self._name



# Expose the latest version as core Jobspec
# If the user wants an earlier version (when we have them)
# they can import it
import jobspec.schema as schemas

from .core import Attributes
from .core import Jobspec as JobspecBase
from .core import Requires, Resources


class Jobspec(JobspecBase):
    schema = schemas.jobspec_nextgen



from .registry import TransformerRegistry

# Only do this once
registry = None


def get_transformer_registry():
    global registry
    if registry is None:
        registry = TransformerRegistry()
        registry.discover()
    return registry



__version__ = "0.1.14"
AUTHOR = "Vanessa Sochat"
AUTHOR_EMAIL = "vsoch@users.noreply.github.com"
NAME = "jobspec"
PACKAGE_URL = "https://github.com/compspec/jobspec"
KEYWORDS = "cluster, orchestration, transformer, jobspec, flux"
DESCRIPTION = "Jobspec specification and translation layer for cluster work"
LICENSE = "LICENSE"


################################################################################
# Global requirements

INSTALL_REQUIRES = (("jsonschema", {"min_version": None}),)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)
INSTALL_REQUIRES_ALL = INSTALL_REQUIRES + TESTS_REQUIRES



#!/usr/bin/env python

import argparse
import os
import sys

import jobspec
from jobspec.logger import setup_logger


def get_parser():
    parser = argparse.ArgumentParser(
        description="Jobspec",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Global Variables
    parser.add_argument(
        "--debug",
        dest="debug",
        help="use verbose logging to debug.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--quiet",
        dest="quiet",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--version",
        dest="version",
        help="show software version.",
        default=False,
        action="store_true",
    )

    subparsers = parser.add_subparsers(
        help="actions",
        title="actions",
        description="actions",
        dest="command",
    )
    subparsers.add_parser("version", description="show software version")

    # Maybe this warrants a better name, but this seems to be what we'd want to do -
    # run a jobspec
    run = subparsers.add_parser(
        "run",
        formatter_class=argparse.RawTextHelpFormatter,
        description="receive and run a jobpsec",
    )

    # This will do the subsystem match before the run
    # run.add_argument("--subsystem-dir", help="directory with subsystem metadata to load")
    run.add_argument("-t", "--transform", help="transformer to use", default="flux")

    # This does just the user space subsystem match
    satisfy = subparsers.add_parser(
        "satisfy",
        formatter_class=argparse.RawTextHelpFormatter,
        description="determine if a jobspec is satisfied by the user subsystem",
    )
    satisfy.add_argument(
        "--subsystem-dir", dest="sdir", help="subsystem directory with JGF to load"
    )

    # If this is True, we do not allow a satisfy to occur if subsystem metadata is entirely missing
    # and the jobspec declares it needed
    satisfy.add_argument(
        "--require-all",
        dest="require_all",
        help="require all subsystems to be present and satisfied.",
        default=False,
        action="store_true",
    )

    for cmd in [run, satisfy]:
        cmd.add_argument("jobspec", help="jobspec yaml file", default="jobspec.yaml")

    return parser


def run_jobspec():
    """
    this is the main entrypoint.
    """
    parser = get_parser()

    def help(return_code=0):
        """print help, including the software version and active client
        and exit with return code.
        """
        version = jobspec.__version__

        print("\nJobspec v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    if args.debug is True:
        os.environ["MESSAGELEVEL"] = "DEBUG"

    # Show the version and exit
    if args.command == "version" or args.version:
        print(jobspec.__version__)
        sys.exit(0)

    setup_logger(
        quiet=args.quiet,
        debug=args.debug,
    )

    # Here we can assume instantiated to get args
    if args.command == "run":
        from .run import main
    elif args.command == "satisfy":
        from .satisfy import main
    else:
        help(1)
    main(args, extra)


if __name__ == "__main__":
    run_jobspec()



import os

from .subsystem import SubsystemRegistry


def get_subsystem_registry(path=None):
    """
    If a path is not defined, we look in the user home at:

    ~/.compspec/subsystems
    """
    if path is None:
        path = os.path.join(os.path.expanduser("~"), ".compspec", "subsystems")

    if not os.path.exists(path):
        raise ValueError(f"User subsystem directory {path} does not exist")

    # Generate the subsystem registry
    return SubsystemRegistry(path)



import os
import stat

import jobspec.utils as utils

from .base import StepBase


class WriterStep(StepBase):
    """
    A writer step has support for writing files

    This step also provides a base for other writing steps.
    """

    name = "write"
    required = ["filename"]

    def validate(self):
        """
        A write step requires a filename, and matching file in scripts
        """
        filename = self.options.get("filename")

        # This validates unique names
        names = list(self.scripts.keys())

        # 1. All filename directives must be defined in scripts
        if filename and filename not in names:
            raise ValueError(
                f"Filename {filename} is defined for a {self.name} step but not in task scripts"
            )

    def run(self, *args, **kwargs):
        """
        write some content to a filename.

        This is currently not used, as we can represent the same logic in a task.

        step: write
        filename: install.sh
        executable: true
        """
        filename = self.options.get("filename")
        stage = self.options.get("stage")
        fullpath = os.path.join(stage, filename)
        utils.write_file(self.scripts[filename]["content"], fullpath)

        executable = self.options.get("executable") is True
        # Execute / search permissions for the user and others
        if executable:
            st = os.stat(fullpath)
            os.chmod(fullpath, st.st_mode | stat.S_IEXEC | stat.S_IXGRP | stat.S_IXOTH)



#!/usr/bin/env python

import os
import sys


def main(args, _):
    """
    Run an extraction. This can be converted to a proper function
    if needed.
    """
    from jobspec.plugin import get_transformer_registry

    registry = get_transformer_registry()

    # This raises an error if not found
    # This would be what we put in a Python script
    # that is in a cronjob, for loop receiver, etc.
    # We can add additional options to the init here
    plugin = registry.get_plugin(args.transform)()

    # The jobspec needs to exist as a file here
    if not os.path.exists(args.jobspec):
        sys.exit(f"JobSpec {args.jobspec} does not exist.")

    # Run the plugin with the jobspec
    plugin.run(args.jobspec)



import shlex


def new_simple_jobspec(command, nodes=1, name=None, tasks=1, jobspec_version=1):
    """
    Generate a new simple jobspec from basic parameters.
    """
    if isinstance(command, str):
        command = shlex.split(command)

    if not command:
        raise ValueError("A command must be provided.")

    # If we don't have a name, derive one
    if name is None:
        name = command[0]

    if nodes < 1 or tasks < 1:
        raise ValueError("Nodes and tasks for the job must be >= 1")

    # Replicas identifies the slot
    rack_resource = {
        "type": "rack",
        "replicas": 1,
        "with": [
            {
                "type": "node",
                "count": nodes,
                "with": [
                    {
                        "type": "core",
                        "count": tasks,
                    }
                ],
            }
        ],
    }

    tasks_resources = [
        {
            "command": command,
            "resources": name,
        }
    ]
    return {
        "version": jobspec_version,
        "resources": {name: rack_resource},
        "tasks": tasks_resources,
    }



import copy


def find_resources(flat, resource, slot, last_one=False):
    """
    Unwrap a nested resource
    """
    # We found a dominant subsystem resource
    # TODO convert flux submit into a jobspec too?
    if "type" in resource and resource["type"] != "slot":
        if "count" not in resource and "replicas" not in resource:
            raise ValueError("A resource must have a count (non-slot) or replicas (slot)")
        flat[resource["type"]] = resource.get("count") or resource.get("replicas")

    # The previous was the found slot, return
    if last_one:
        return True

    # We found the slot, this is where we stop
    if "type" in resource and resource["type"] == "slot":
        last_one = True

    # More traversing...
    if "with" in resource:
        for r in resource["with"]:
            find_resources(flat, r, slot, last_one)
    return flat


def to_jobspec(resource, js=None, slot_name=None, has_slot=False):
    """
    Recursive function to help convert to jobspec

    There could be more than one slot in the future, but in
    practice now most flux jobspecs just support one.
    """
    slot_name = slot_name or "default"

    # If we don't have flat yet, make it the entire thing
    if not js:
        js = copy.deepcopy(resource)
    else:
        if "with" not in js:
            js["with"] = []

    # We found a place to insert a slot
    if "replicas" in resource:
        has_slot = True
        if "replicas" not in resource:
            raise ValueError("A resource must have a count (non-slot) or replicas (slot)")

        # If we have a slot, expand out into one
        count = resource["replicas"]
        with_list = resource.get("with") or []
        slot = {"type": "slot", "count": count, "label": slot_name}
        if with_list:
            slot["with"] = with_list
            resource["with"] = [slot]

    # If count not in resources, assume 1
    if "count" not in resource:
        resource["count"] = 1

    # Delete requires, it isn't understood
    if "requires" in resource:
        del resource["requires"]
    if "replicas" in resource:
        del resource["replicas"]

    # More traversing...
    if "with" in resource:
        for r in resource["with"]:
            has_slot = to_jobspec(r, js, slot_name, has_slot)
    return has_slot


def parse_resource_subset(named_resources, resources):
    """
    Parse and validate the resource subset.

    Note that right now we rely on the user to ask for sensical values.
    For example, a task in a group (batch) that asks for more GPU than
    the batch has won't be satisfied. But if this is a grow/autoscale
    setup, maybe it eventually could be, so we allow it.
    """
    # If we are given a Resources object, unwrap the data
    if hasattr(resources, "data"):
        resources = resources.data
    if hasattr(named_resources, "data"):
        named_resources = named_resources.data

    # Case 1: we have resources as a string and it's a member of named
    if isinstance(resources, str):
        if "|" in resources:
            raise ValueError("Asking for an OR in resources is not supported yet.")
        if "," in resources:
            raise ValueError("Asking for an AND in resources is not supported yet.")
        if resources not in named_resources:
            raise ValueError(f"Asked for resources '{resources}' that are not known")
        return named_resources[resources]

    # Case 2: It's just it's own thing
    return resources



import copy

import jobspec.core as js
import jobspec.core.resources as rcore
from jobspec.logger import LogColors
from jobspec.runner import TransformerBase

from .steps import batch, stage, submit


class FluxWorkload(TransformerBase):
    """
    The flux workload (batch and submit) transformer

    The transformer needs to share state between tasks and groups, so we create
    a shared class to do that. A "flux workload" is some number of tasks / batches
    It replaces the step abstractions to be run.
    """

    # These metadata fields are required (and checked for)
    name = "flux"
    description = "Flux Framework workload"

    def announce(self):
        """
        Announce prints an additional prefix during run
        """
        prefix = "flux workload".ljust(15)
        print(f"=> {LogColors.OKCYAN}{prefix}{LogColors.ENDC}")

    def parse(self, jobspec):
        """
        Parse the jobspec into tasks for flux.
        """
        # Reset the jobspec and groups and tasks
        self.js = jobspec
        self.group_lookup = {}

        # Top level, a-la-carte tasks (steps)
        self.tasks = []

        # Parse top level groups into a lookup. Those that don't have a name
        # are given a name based on order, and assumed not to be linked to anything
        for i, group in enumerate(self.js.get("groups") or []):
            name = group.get("name") or f"batch-{i}"
            self.group_lookup[name] = group

        # We will return a listing of steps to complete - flux submit each
        # A group referenced within a task group is parsed there
        tasks = self.js.get("tasks") or []
        if tasks:
            self.tasks = self.parse_tasks(tasks, self.resources, requires=self.requires)

        # Now parse remaining groups "a la carte" that aren't linked to top level tasks
        # We copy because otherwise the dict changes size when the task parser removes
        groups = copy.deepcopy(self.group_lookup)
        for name, group in groups.items():
            # Check if the group was already removed by another group task,
            # and don't run if it was!
            if name not in self.group_lookup:
                continue

            self.tasks.insert(
                0, self.parse_group(group, name, self.resources, requires=self.requires)
            )

        # Return the transformer to call run to
        return self.tasks

    @property
    def resources(self):
        """
        This returns a global resource lookup
        """
        return self.js.get("resources", {})

    @property
    def requires(self):
        return self.js.get("requires", {})

    def parse_group(self, group, name, resources=None, requires=None, attributes=None):
        """
        Parse a group and return a step. If tasks are within a group,
        they are written to the script of the group.

        This might be tricky to figure out for the very nested version.
        """
        resources = resources or {}
        requires = requires or {}
        attributes = attributes or {}

        group_resources = group.get("resources", {})
        group_attributes = group.get("attributes", {})

        # Group requires gets updated from globa (and passed to task)
        group_requires = js.Requires(requires)
        group_requires.update(group.get("requires"))

        # Group resources don't have a slot
        group_resources = js.Resources(rcore.parse_resource_subset(resources, group_resources))

        # Parse the task steps for the group
        tasks = group.get("tasks") or []
        steps = []
        if tasks:
            name_prefix = f"{name}-"
            steps = self.parse_tasks(
                tasks,
                # Still provide global resources as a lookup
                resources=resources,
                requires=group_requires.data,
                attributes=group_attributes,
                name_prefix=name_prefix,
            )

        # Prepare a batch (group) step
        new_step = batch(
            self.js,
            name=name,
            resources=group_resources,
            attributes=group_attributes,
            requires=group_requires,
            tasks=tasks,
        )
        new_step.tasks = steps
        return new_step

    def parse_task(self, task, name, resources=None, requires=None, attributes=None):
        """
        Parse a task and return a step.
        """
        task_resources = task.get("resources", {})

        # The slot is optional and drives where the match is targeting
        slot = task.get("slot")

        # If the task has resources, must be:
        #  A named section in the global resources
        #  A subset of parent resources
        task_resources = js.Resources(
            rcore.parse_resource_subset(resources, task_resources), slot=slot
        )

        # Derive and update task attributes, if provided
        task_attributes = js.Attributes(attributes).update(task.get("attributes"))

        # Same for requires. This might eventually include retrieval of
        # artifact metadata first.
        task_requires = js.Requires(requires).update(task.get("requires"))

        # Prepare a submit step
        return submit(
            self.js,
            name=name,
            resources=task_resources,
            attributes=task_attributes,
            requires=task_requires,
            task=task,
        )

    def parse_tasks(self, tasks, resources=None, attributes=None, requires=None, name_prefix=None):
        """
        Parse a jobspec (or group) tasks and return a list of steps.

        If the task is defined in a group, the attributes provided will be inherited
        from the group. Otherwise they will likely be empty.
        """
        # A name prefix helps with defaults to scope to a group
        name_prefix = name_prefix or ""
        resources = resources or {}
        attributes = attributes or {}
        requires = requires or {}

        steps = []
        for i, task in enumerate(tasks):
            # Create a name based on the index or the task name
            name = task.get("name") or f"{name_prefix}task-{i}"

            # Case 1: We found a group! Parse here and add to steps
            group_name = task.get("group")
            if group_name is not None:
                group = self.group_lookup.get(group_name)
                if not group:
                    raise ValueError(
                        f"Task {name} is looking for group {group_name} that is not defined."
                    )

                # We assume a group is used once
                del self.group_lookup[group_name]
                new_step = self.parse_group(
                    group, group_name, resources, requires=requires, attributes=attributes
                )

            # Case 2: we have a regular task to flux submit
            else:
                new_step = self.parse_task(
                    task, name, resources, requires=requires, attributes=attributes
                )

            steps.append(new_step)

        return steps


# A transformer can register shared steps, or custom steps
FluxWorkload.register_step(batch)
FluxWorkload.register_step(submit)
FluxWorkload.register_step(stage)



class StepBase:
    """
    A base step describes the design of a step.
    """

    required = []

    def __init__(self, js, **kwargs):
        """
        A step takes a task definition and custom options

        Note that we aren't using options now, but could be.
        """
        self.jobspec = js
        if not hasattr(self, "name"):
            raise ValueError(f"Step {self} is missing a name")

        # Each step can take custom options (the keyword arguments)
        self.options = kwargs

        # Shared validation
        self._validate()

        # Custom validation and setup
        self.validate()
        self.setup(**kwargs)

    def _validate(self):
        """
        Shared validation functions
        """
        # Required fields are all... required.
        for field in self.required:
            if field not in self.options:
                raise ValueError(f"Step {self.name} is missing field {field}")
            if self.options[field] is None:
                raise ValueError(f"Step {self.name} has undefined field {field}")

    def validate(self):
        """
        Validate the step
        """
        pass

    def setup(self, **kwargs):
        """
        Custom setup

        Akin to overriding init, but easier to write.
        """
        pass

    def run(self, *args, **kwargs):
        """
        Run a step.

        This is the argument structure that should be used.
        """
        raise NotImplementedError



# This imports the latest version
import jobspec.core as js
import jobspec.steps.runner as step_runner


class TransformerBase:
    """
    A Transformer base is the core of a JobSpec

    It loads a Jobspec and transforms for a particular environment
    (which typically means a workload manager or similar). This is most
    relevant for submit and batch commands, along with custom steps.
    """

    steps = {}

    def __init__(self, **options):
        """
        Create a new transformer backend, accepting any options type.

        Validation of transformers is done by the registry
        """
        # Set options as attributes
        for key, value in options.items():
            setattr(self, key, value)

    @classmethod
    def register_step(cls, step, name=None):
        """
        Register a step class to the transformer
        """
        # Allow registering an empty step if needed
        # An empty step does nothing, an explicit declaration
        # by the transformer developer it's not needed, etc.
        name = name or step.name
        cls.steps[name] = step

    def parse(self, jobspec):
        """
        parse validates logic and returns a series of steps, which
        will depend on the underlying transformer. It might be:

        - a batch that includes logic to write a script and then run it
        - one or more submit commands
        """
        raise NotImplementedError

    def announce(self):
        pass

    def flatten(self, filename):
        raise NotImplementedError

    def run(self, filename):
        """
        Run the transformer
        """
        # Load the jobspec
        jobspec = self.load_jobspec(filename)

        # Get validated transformation steps
        # These will depend on the transformer logic
        steps = self.parse(jobspec)
        self.announce()

        # Run each step to submit the job, and that's it.
        for step in steps:
            step_runner.run(self.name, step)

    def load_jobspec(self, filename):
        """
        Load and transform a jobspec.

        This function should be able to load it in some raw format
        and convert into correct directives given the transformer.
        """
        return js.Jobspec(filename)



import os
import stat

import jobspec.utils as utils

from .base import StepBase


class EmptyStep(StepBase):
    """
    An empty step is used to declare that a step should be skipped
    """

    def run(self, *args, **kwargs):
        """
        do nothing.
        """
        pass



import copy

import jsonschema

import jobspec.schema as schema
from jobspec.logger.generate import generate_name

from .base import ResourceBase
from .resources import find_resources, to_jobspec


class Jobspec(ResourceBase):
    def __init__(self, filename, validate=True, name=None, schema=schema.jobspec_nextgen):
        """
        Load in and validate a Jobspec
        """
        self.name = name or generate_name()

        # This should typically be loaded from jobspec.core
        if not hasattr(self, "schema") or not self.schema:
            self.schema = schema
        self.data = None
        self.load(filename)
        if validate:
            self.validate()

    def validate(self):
        """
        Validate the jsonschema
        """
        jsonschema.validate(self.data, self.schema)

        # Require at least one of command or steps, unless it is a group
        for task in self.data.get("tasks", []):
            if "group" not in task and ("command" not in task and "steps" not in task):
                raise ValueError("Jobspec is not valid, each task must have a command or steps")


class Resources(ResourceBase):
    def __init__(self, data, slot=None):
        """
        Interact with loaded resources.
        """
        self.data = data
        self.slot = slot

    def flatten_slot(self, slot=None):
        """
        Find the task slot, flatten it, and return
        """
        slot = slot or self.slot

        # Traverse each section. There is usually only one I guess
        flat = {}
        find_resources(flat, self.data, slot)
        return flat

    def to_jobspec(self, slot_name):
        """
        Turn the resource into a flux jobspec

        Note this is not currently used - it was too error prone
        """
        slot = self.slot or {}
        label = slot_name or (slot.get("label") or "default")

        # Hold onto the original data so it is not mangled
        js = copy.deepcopy(self.data)

        # Traverse each section and convert to flux jobspec
        has_slot = to_jobspec(js, slot_name=label)

        # If we don't have a slot, we have to make a fake one at the top
        if not has_slot:
            js = {"type": "slot", "count": 1, "label": label, "with": [js]}
        return js


class Attributes(ResourceBase):
    """
    Job attributes, not formally defined yet.
    """

    pass


class Requires(ResourceBase):
    """
    Requires are nested groups
    """

    def update(self, requires):
        """
        Update specific groups. This is assumed
        at the level of the attribute, not the group.
        E.g., this at the global level:

        requires:
          io:
            fieldA: valueA
            fieldB: valueB

        Updated with this:
        requires:
          io:
            fieldB: valueC

        Results in this:
        requires:
          io:
            fieldA: valueA
            fieldB: valueC
        """
        if not requires:
            return
        for group, fields in requires.items():
            # If we don't have the group at all, we can add all and continue!
            if group not in self.data:
                self.data[group] = fields
                continue

            # If we have the group, update on the level of fields
            self.data[group].update(fields)
        return self



import sys

from jobspec.logger import LogColors


def run(name, step):
    """
    Run a single step. Make it pretty.
    """
    prefix = f"{name} {step.name}".ljust(15)
    print(f"=> {LogColors.OKCYAN}{prefix}{LogColors.ENDC}", end="")
    try:
        result = step.run()
        if not result:
            return
        print(
            f"{LogColors.OKBLUE}{result.out}{LogColors.ENDC} {LogColors.OKGREEN}OK{LogColors.ENDC}"
        )
        result.print_extra()
    except Exception as e:
        print(f"\n{LogColors.RED}{str(e)}{LogColors.ENDC}")
        sys.exit()



create_subsystem_sql = """
CREATE TABLE subsystems (
  name TEXT PRIMARY KEY NOT NULL
);"""

create_nodes_sql = """CREATE TABLE nodes (
  subsystem TEXT NOT NULL,
  label TEXT NOT NULL,
  type TEXT NOT NULL,
  basename TEXT NOT NULL,
  name TEXT NOT NULL,
  id INTEGER NOT NULL,
  FOREIGN KEY(subsystem) REFERENCES subsystems(name),
  UNIQUE (subsystem, label)
);"""

create_attributes_sql = """CREATE TABLE attributes (
  name TEXT NOT NULL,
  value TEXT NOT NULL,
  node TEXT NOT NULL,
  FOREIGN KEY(node) REFERENCES nodes(label)
);"""



import copy
import json
import os
import uuid

import jobspec.utils as utils
from jobspec.steps.base import StepBase
from jobspec.transformer.result import Result

script_prefix = ["#!/bin/bash"]

# Custom Flux steps - just write and register!


class stage(StepBase):
    """
    A copy step uses flux filemap to stage across nodes

    This assumes we don't have a shared filesystem. It is skipped if we do.
    """

    name = "stage"

    def run(self, stage, *args, **kwargs):
        """
        Run the stage step = fall back to filename for now
        """
        # If we have a sharedfs, return early, the write will have written it there
        sharedfs = self.options.get("settings", {}).get("sharedfs") is True
        if sharedfs:
            return

        # Sanity check staging directory exists across nodes
        cmd = ["flux", "exec", "-r", "all", "-x", "0", "mkdir", "-p", stage]
        utils.run_command(cmd, check_output=True)

        name = str(uuid.uuid4())
        filename = self.options["filename"]
        cmd = ["flux", "filemap", "map", "--tags", name, "--directory", stage, filename]
        utils.run_command(cmd, check_output=True)

        # Assume we send to all ranks besides where we've already written it
        # This will likely fail if the filesystem is shared
        cmd = [
            "flux",
            "exec",
            "--dir",
            stage,
            "-r",
            "all",
            "-x",
            "0",
            "flux",
            "filemap",
            "get",
            "--tags",
            name,
        ]
        utils.run_command(cmd, check_output=False)

        # Unmap to clear the memory map
        cmd = ["flux", "filemap", "unmap", "--tags", name]
        utils.run_command(cmd, check_output=True)


class JobBase(StepBase):
    """
    base with shared logic for submit or batch
    """

    def cleanup(self, filename):
        """
        Cleanup a filename if it still exists
        """
        if os.path.exists(filename):
            os.remove(filename)

    def prepare(self, command=None, waitable=False):
        """
        Return the command, without flux submit|batch
        """
        cmd = []

        # We can get the resources from options
        resources = self.options.get("resources")

        # These aren't used yet - they need to go into flux
        attributes = self.options.get("attributes") or {}
        task = self.options.get("task") or {}

        # This flattens to be what we ask flux for
        slot = resources.flatten_slot()
        nodes = slot.get("node")
        tasks = slot.get("core")
        gpus = slot.get("gpu")

        # Get name, jobspec, depends, etc
        name = self.options.get("name")
        duration = attributes.get("duration")
        cwd = attributes.get("cwd")
        watch = attributes.get("watch")

        # Note that you need to install our frobnicator plugin
        # for this to work. See the examples/depends_on directory
        for depends_on in task.get("depends_on") or []:
            cmd += [f"--setattr=dependency.name={depends_on}"]

        if cwd is not None:
            cmd += ["--cwd", cwd]
        if name is not None:
            cmd += ["--job-name", name]
        if duration is not None:
            cmd += ["--time-limit", str(duration)]
        if watch is True:
            cmd += ["--watch"]

        if nodes:
            cmd += ["-N", str(nodes)]
        if tasks:
            cmd += ["-n", str(tasks)]
        if gpus:
            cmd += ["-g", str(gpus)]

        # Replicas we do with cc
        replicas = task.get("replicas")
        if replicas:
            replicas -= 1
            cmd += ["--cc", f"0-{replicas}"]

        # Are we adding a waitable flag? This will prevent the batch
        # from exiting
        if waitable:
            cmd += ["--flags=waitable"]

        # Right now assume command is required
        if not command:
            command = task["command"]
        if isinstance(command, str):
            command = [command]
        cmd += command
        return cmd


class batch(JobBase):
    """
    flux batch is the flux implementation of a group.

    It starts an allocation to run commands and launch
    other tasks or groups.
    """

    name = "batch"

    def setup(self, **kwargs):
        """
        Custom setup

        A batch can hold one or more task steps, and these need to be written into
        a custom script (and not run as tasks)
        """
        self.tasks = []

    def write_tasks_script(self):
        """
        Generate the batch script.
        """
        data = copy.deepcopy(script_prefix)
        for task in self.tasks:
            if task.name == "batch":
                cmd, _ = task.generate_command(waitable=True)
                data.append(" ".join(cmd))
                # This is the jobspec
                data.append(f"# rm -rf {cmd[-1]}")
            else:
                data.append(" ".join(task.generate_command(waitable=True)))

        # Ensure all jobs are waited on
        data.append("flux job wait --all")
        script = "\n".join(data)
        return {"mode": 33216, "data": script, "encoding": "utf-8"}

    def generate_command(self, waitable=False):
        """
        Convenience function to generate the command.

        This is also intended for flux batch to use,
        and we expect the last argument to be the temporary file
        that needs to be cleaned up
        """
        # With batch, we are going to cheat and run a flux job submit --dry-run
        # with the command that would normally be derived for flux batch.
        command = ["flux", "broker", "{{tmpdir}}/batch-script"]
        cmd = self.prepare(command)

        # This returns a partial jobspec for the work that is needed...
        cmd = ["flux", "submit", "--dry-run"] + cmd
        res = utils.run_command(cmd, check_output=True)

        if res["return_code"] != 0:
            raise ValueError(f"Issue generating flux jobspec: {res['message']}")

        # Then we will add our (not written) batch script to the files section
        # Load the jobspec and add our "files" section to it with tasks (flux submit, etc)
        js = json.loads(res["message"])

        # Be careful about updating files - not sure if there are cases
        # when there might be other content there.
        files = js["attributes"]["system"].get("files") or {}
        files["batch-script"] = self.write_tasks_script()
        js["attributes"]["system"]["files"] = files

        # Write the jobspec to a temporary file, target for cleanup
        tmpfile = utils.get_tmpfile(prefix="jobspec-")
        utils.write_file(json.dumps(js), tmpfile)

        # Prepare a result, we can show the batch script if debug is on
        result = Result()
        for line in files["batch-script"]["data"].split("\n"):
            result.add_debug_line(line)

        # and submit with flux job submit <jobspec>
        # We can't really watch here, or do anything with attributes yet
        cmd = ["flux", "job", "submit"]
        if waitable:
            cmd += ["--flags=waitable"]
        cmd.append(tmpfile)
        return cmd, result

    def run(self, *args, **kwargs):
        """
        Run the batch step
        """
        cmd, result = self.generate_command()
        res = utils.run_command(cmd, check_output=True)

        # The temporary file to cleanup is the last in the list
        tmpfile = cmd[-1]

        # Prepare a result to return
        result.out = res["message"].strip()
        result.add_debug_line(" ".join(cmd))

        # Cleanup the files
        self.cleanup(tmpfile)
        return result


class submit(JobBase):
    name = "submit"

    def generate_command(self, waitable=False):
        """
        Convenience function to generate the command.

        This is intended for flux batch to use
        """
        cmd = self.prepare(waitable=waitable)
        return ["flux", "submit"] + cmd

    def run(self, *args, **kwargs):
        """
        Run the submit step.

        The python bindings are giving me weird errors.
        """
        cmd = self.generate_command()

        # Are we watching?
        attributes = self.options.get("attributes") or {}
        watch = attributes.get("watch")
        res = utils.run_command(cmd, check_output=True, stream=watch)

        # Prepare a result to return
        result = Result()

        # Return results to print
        if not watch:
            result.out = res["message"].strip()
        result.add_debug_line(" ".join(cmd))
        return result
